import 'dart:io';

import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extention/extension.dart';
import 'package:permission_handler/permission_handler.dart';

part 'module_ble.g.dart';

@CopyWith()
class ModuleBle {
  ModuleBle({
    required UUID this.uuid,
    this.scanFilter = Builtin.scanFilterWearEcg,
    this.setDevice,
    this.isSimulator = false,
  });

  BluetoothDevice? device;

  final bool isSimulator;
  bool isSimulatorConnected = false;

  final void Function(BluetoothDevice device)? setDevice;

  final bool Function(BluetoothDevice device) scanFilter;

  // bool Function(ScanResult scanResult) get _scanResultFilter =>
  //     (scanResult) => scanFilter(scanResult.device);
  //
  // int _sortByRssi(a, b) => b.rssi - a.rssi;

  DateTime? _sessionTime;
  final streams = StreamInterface();

  final UUID uuid;
  final points = <Enum, BluetoothCharacteristic>{};

  Future<void> _requestPermission() async {
    if (Platform.isWindows) return;
    final permissions = [
      Permission.bluetoothScan,
      Permission.bluetooth,
      Permission.bluetoothConnect,
      Permission.bluetoothAdvertise,
    ];
    await permissions.request();
    'Permissions are request: $permissions'.log;
  }

  Future<bool> connectAndRegister([
    String preferredDeviceId = '',
    bool strictConnection = false,
  ]) async {
    final isConnected = await connectDevice(
      preferredDeviceId,
      strictConnection,
    );

    if (isConnected) await registerCharacteristics();

    return isConnected;
  }

  Future<List<BluetoothDevice>> findConnectedDevices({
    bool isFiltered = true,
  }) async {
    await _requestPermission();

    final connected = await FlutterBluePlus.systemDevices;
    if (isFiltered) {
      return connected.where(scanFilter).toList();
    }
    return connected;
  }

  Future<List<ScanResult>> findScannedDevices({
    bool isFiltered = true,
  }) async {
    await _requestPermission();

    final scanned = <ScanResult>[];
    Set<DeviceIdentifier> seen = {};

    final subscription = FlutterBluePlus.scanResults.listen((results) {
      for (final r in results) {
        if (!seen.contains(r.device.remoteId)) {
          seen.add(r.device.remoteId);
          scanned.add(r);
        }
      }
    });

    await FlutterBluePlus.startScan();

    await Future.delayed(const Duration(seconds: 2));

    await FlutterBluePlus.stopScan();
    subscription.cancel();

    if (isFiltered) {
      return scanned.where((r) => scanFilter(r.device)).toList();
    }
    return scanned;
  }

  Future<bool> connectDevice([
    String preferredDeviceId = '',
    bool strictConnection = false,
  ]) async {
    if (isSimulator) {
      isSimulatorConnected = true;
      return true;
    }

    /// 0. initialize
    device = null;
    'Device is set to $device'.log;

    /// 1. devices which are connected
    final connected = await findConnectedDevices();

    device = connected
        .where((device) => device.remoteId.toString() == preferredDeviceId)
        .firstOrNull;
    if (device != null) return true;

    if (!strictConnection && device == null) {
      device = connected.firstOrNull;
    }
    if (device != null) return true;

    /// 2. perform scanning process
    final scanned = await findScannedDevices();

    device = scanned
        .where((r) => r.device.remoteId.toString() == preferredDeviceId)
        .firstOrNull
        ?.device;
    if (device != null) {
      await device?.connect();
      'Device named ${device?.platformName} has been connected now'.log;
      setDevice?.call(device!);
      return true;
    }

    if (!strictConnection && device == null) {
      device = scanned.firstOrNull?.device;
    }

    if (device != null) {
      await device?.connect();
      'Device named ${device?.platformName} has been connected now'.log;
      setDevice?.call(device!);
      return true;
    }

    /// 3. fail to connect device
    return false;
  }

  Future<void> disconnectDevice({
    BluetoothDevice? device,
  }) async {
    /// simulator
    if (isSimulator) {
      isSimulatorConnected = false;
      return;
    }

    if (device != null) {
      await device.disconnect();
      'Device named ${device.platformName} has been disconnected now'.log;
      return;
    }

    final connected = await FlutterBluePlus.systemDevices;
    'connected:\n$connected'.log;

    await Future.wait(
      [
        for (final device in connected.where(scanFilter))
          () async {
            await device.disconnect();
            'Device named ${device.platformName} has been disconnected now'.log;
          }(),
      ],
    );
    device = null;
  }

  void _setupBleStreams() {
    _sessionTime = DateTime.now();

    streams
      ..connection = _getConnectionStateStream().broadcast
      ..battery = _getBatteryLevelStream().broadcast
      ..tap = _getPhysicalTapStream().broadcast;
  }

  Future<void> registerCharacteristics() async {
    /// simulator
    if (isSimulator) return _setupBleStreams();

    /// Non-simulator
    if (device == null) return;

    final services = await device!.discoverServices();
    for (final service in services) {
      final characteristics = service.characteristics;
      for (final characteristic in characteristics) {
        uuid.characteristic.forEach(
          (key, value) {
            final uuid = characteristic.uuid.toString();
            if (value == null) return;
            if (uuid.toLowerCase() == value.toLowerCase()) {
              '${key.name} was registered [${uuid.toLowerCase()}]'.log;
              points[key] = characteristic;
            }
          },
        );
      }
    }
    _setupBleStreams();
  }

  Stream<bool?> _getConnectionStateStream() async* {
    /// simulator
    if (isSimulator) {
      final connectionStream = Stream.periodic(
        const Duration(seconds: 2),
        (_) => isSimulatorConnected,
      );
      await for (final event in connectionStream) {
        yield event;
      }
      return;
    }

    final stream = device?.connectionState.broadcast;
    if (stream == null) return;

    var previousEmittedTime = DateTime.now();
    BluetoothConnectionState? prev;

    final subscription = stream.listen((event) {
      prev = event;
      previousEmittedTime = DateTime(0);
    });

    while (device != null) {
      final timeAfterLastEmit = DateTime.now().difference(previousEmittedTime);
      if (timeAfterLastEmit > const Duration(seconds: 2)) {
        yield prev == BluetoothConnectionState.connected;
        previousEmittedTime = DateTime.now();
      }
      await Future<void>.delayed(const Duration(seconds: 2));
    }

    yield null;

    await subscription.cancel();
  }

  Stream<int?> _getBatteryLevelStream() async* {
    /// simulator
    if (isSimulator) {
      final batteryStream = Stream.periodic(
        const Duration(seconds: 2),
        (_) => isSimulatorConnected ? 100 : null,
      );
      await for (final event in batteryStream) {
        yield event;
      }
      return;
    }

    final battery = points[Characteristics.BATTERY_LEVEL];
    if (battery == null) return;

    try {
      await battery.setNotifyValue(true);
    } catch (e) {
      '_getBatteryLevelStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await battery.setNotifyValue(true);
      } catch (e) {
        '_getBatteryLevelStream (second try): $e'.log;
      }
    }

    final read = await battery.read();
    var prev = read.isNotEmpty ? read[0] : null;
    yield prev;

    var previousEmitTime = DateTime.now();
    // final subscription = battery.lastValueStream.listen((event) {
    //   final changed = event.isNotEmpty ? event[0] : null;
    //   prev = changed;
    //   previousEmitTime = DateTime(0);
    // });

    while (device != null) {
      final timeAfterLastEmit = DateTime.now().difference(previousEmitTime);
      if (timeAfterLastEmit > const Duration(seconds: 5)) {
        final read = await battery.read();
        prev = read.isNotEmpty ? read[0] : null;
        yield prev;
        previousEmitTime = DateTime.now();
      }
      await Future<void>.delayed(const Duration(seconds: 5));
    }
    yield null;
  }

  Stream<TapEvent?> _getPhysicalTapStream() async* {
    /// simulator
    if (isSimulator) return;

    final tap = points[Characteristics.TAP_BUTTON];
    if (tap == null) return;
    final sessionTimeInside = _sessionTime;

    try {
      await tap.setNotifyValue(true);
    } catch (e) {
      '_getPhysicalTapStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await tap.setNotifyValue(true);
      } catch (e) {
        '_getPhysicalTapStream (second try): $e'.log;
      }
    }

    await for (final event in tap.lastValueStream) {
      if (sessionTimeInside != _sessionTime) break;
      if (event.isEmpty) continue;
      yield TapEvent(time: DateTime.now());
    }
    yield null;
  }
}
